from sqlframe.base.mixins.table_mixins import (
    DeleteSupportMixin,
    MergeSupportMixin,
    UpdateSupportMixin,
)
from sqlframe.base.table import (
    WhenMatched,
    WhenNotMatched,
    WhenNotMatchedBySource,
    _BaseTable,
)
from sqlframe.postgres.dataframe import PostgresDataFrame


class PostgresTable(
    PostgresDataFrame,
    UpdateSupportMixin["PostgresDataFrame"],
    DeleteSupportMixin["PostgresDataFrame"],
    MergeSupportMixin["PostgresDataFrame"],
    _BaseTable["PostgresDataFrame"],
):
    _df = PostgresDataFrame
    _merge_supported_clauses = [WhenMatched, WhenNotMatched, WhenNotMatchedBySource]
    _merge_support_star = False



from __future__ import annotations

import re
import typing as t

from sqlglot import exp
from sqlglot.helper import ensure_list

from sqlframe.base.column import Column

CALLING_CLASS = t.TypeVar("CALLING_CLASS")


def func_metadata(unsupported_engines: t.Optional[t.Union[str, t.List[str]]] = None) -> t.Callable:
    def _metadata(func: t.Callable) -> t.Callable:
        def wrapper(*args, **kwargs):
            funcs_to_not_auto_alias = [
                "posexplode",
                "explode_outer",
                "json_tuple",
                "posexplode_outer",
                "stack",
                "inline",
                "inline_outer",
                "window",
                "session_window",
                "window_time",
            ]

            result = func(*args, **kwargs)
            if (
                isinstance(result, Column)
                and isinstance(result.column_expression, exp.Func)
                and not isinstance(result.expression, exp.Alias)
                and func.__name__ not in funcs_to_not_auto_alias
            ):
                col_name = result.column_expression.find(exp.Identifier)
                if col_name:
                    col_name = col_name.name
                else:
                    col_name = result.column_expression.find(exp.Literal)
                    if col_name:
                        col_name = col_name.this
                alias_name = f"{func.__name__}__{col_name or ''}__"
                # BigQuery has restrictions on alias names so we constrain it to alphanumeric characters and underscores
                return result.alias(re.sub("\W", "_", alias_name))  # type: ignore
            return result

        wrapper.unsupported_engines = (  # type: ignore
            ensure_list(unsupported_engines) if unsupported_engines else []
        )
        return wrapper

    return _metadata



from __future__ import annotations

import logging
import typing as t

from sqlframe.base.dataframe import (
    BaseDataFrame,
    _BaseDataFrameNaFunctions,
    _BaseDataFrameStatFunctions,
)
from sqlframe.base.mixins.dataframe_mixins import (
    NoCachePersistSupportMixin,
    TypedColumnsFromTempViewMixin,
)
from sqlframe.duckdb.group import DuckDBGroupedData

if t.TYPE_CHECKING:
    from sqlframe.duckdb.session import DuckDBSession  # noqa
    from sqlframe.duckdb.readwriter import DuckDBDataFrameWriter  # noqa
    from sqlframe.duckdb.group import DuckDBGroupedData  # noqa
    from pyarrow import Table as ArrowTable, RecordBatchReader

logger = logging.getLogger(__name__)


class DuckDBDataFrameNaFunctions(_BaseDataFrameNaFunctions["DuckDBDataFrame"]):
    pass


class DuckDBDataFrameStatFunctions(_BaseDataFrameStatFunctions["DuckDBDataFrame"]):
    pass


class DuckDBDataFrame(
    NoCachePersistSupportMixin,
    TypedColumnsFromTempViewMixin,
    BaseDataFrame[
        "DuckDBSession",
        "DuckDBDataFrameWriter",
        "DuckDBDataFrameNaFunctions",
        "DuckDBDataFrameStatFunctions",
        "DuckDBGroupedData",
    ],
):
    _na = DuckDBDataFrameNaFunctions
    _stat = DuckDBDataFrameStatFunctions
    _group_data = DuckDBGroupedData

    def explain(
        self, extended: t.Optional[t.Union[bool, str]] = None, mode: t.Optional[str] = None
    ) -> None:
        results = self._get_explain_plan_rows()
        print(results[0][1])

    @t.overload
    def toArrow(self) -> ArrowTable: ...

    @t.overload
    def toArrow(self, batch_size: int) -> RecordBatchReader: ...

    def toArrow(self, batch_size: t.Optional[int] = None) -> t.Union[ArrowTable, RecordBatchReader]:
        self._collect(skip_rows=True)
        if not batch_size:
            return self.session._last_result.arrow()
        return self.session._last_result.fetch_record_batch(batch_size)



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlframe.base.mixins.readwriter_mixins import PandasLoaderMixin, PandasWriterMixin
from sqlframe.base.readerwriter import (
    _BaseDataFrameReader,
    _BaseDataFrameWriter,
)

if t.TYPE_CHECKING:
    from sqlframe.redshift.session import RedshiftSession  # noqa
    from sqlframe.redshift.dataframe import RedshiftDataFrame  # noqa
    from sqlframe.redshift.table import RedshiftTable  # noqa


class RedshiftDataFrameReader(
    PandasLoaderMixin["RedshiftSession", "RedshiftDataFrame"],
    _BaseDataFrameReader["RedshiftSession", "RedshiftDataFrame", "RedshiftTable"],
):
    pass


class RedshiftDataFrameWriter(
    PandasWriterMixin["RedshiftSession", "RedshiftDataFrame"],
    _BaseDataFrameWriter["RedshiftSession", "RedshiftDataFrame"],
):
    pass



from sqlframe.base.mixins.table_mixins import (
    DeleteSupportMixin,
    MergeSupportMixin,
    UpdateSupportMixin,
)
from sqlframe.base.table import (
    WhenMatched,
    WhenNotMatched,
    WhenNotMatchedBySource,
    _BaseTable,
)
from sqlframe.bigquery.dataframe import BigQueryDataFrame


class BigQueryTable(
    BigQueryDataFrame,
    UpdateSupportMixin["BigQueryDataFrame"],
    DeleteSupportMixin["BigQueryDataFrame"],
    MergeSupportMixin["BigQueryDataFrame"],
    _BaseTable["BigQueryDataFrame"],
):
    _df = BigQueryDataFrame
    _merge_supported_clauses = [WhenMatched, WhenNotMatched, WhenNotMatchedBySource]
    _merge_support_star = False



from sqlframe.base.functions import abs as abs
from sqlframe.base.functions import acos as acos
from sqlframe.base.functions import acosh as acosh
from sqlframe.base.functions import add_months as add_months
from sqlframe.base.functions import aes_decrypt as aes_decrypt
from sqlframe.base.functions import aes_encrypt as aes_encrypt
from sqlframe.base.functions import aggregate as aggregate
from sqlframe.base.functions import any_value as any_value
from sqlframe.base.functions import approx_count_distinct as approx_count_distinct
from sqlframe.base.functions import approx_percentile as approx_percentile
from sqlframe.base.functions import approxCountDistinct as approxCountDistinct
from sqlframe.base.functions import array as array
from sqlframe.base.functions import array_agg as array_agg
from sqlframe.base.functions import array_append as array_append
from sqlframe.base.functions import array_compact as array_compact
from sqlframe.base.functions import array_contains as array_contains
from sqlframe.base.functions import array_distinct as array_distinct
from sqlframe.base.functions import array_except as array_except
from sqlframe.base.functions import array_insert as array_insert
from sqlframe.base.functions import array_intersect as array_intersect
from sqlframe.base.functions import array_join as array_join
from sqlframe.base.functions import array_max as array_max
from sqlframe.base.functions import array_min as array_min
from sqlframe.base.functions import array_position as array_position
from sqlframe.base.functions import array_prepend as array_prepend
from sqlframe.base.functions import array_remove as array_remove
from sqlframe.base.functions import array_repeat as array_repeat
from sqlframe.base.functions import array_size as array_size
from sqlframe.base.functions import array_sort as array_sort
from sqlframe.base.functions import array_union as array_union
from sqlframe.base.functions import arrays_overlap as arrays_overlap
from sqlframe.base.functions import arrays_zip as arrays_zip
from sqlframe.base.functions import asc as asc
from sqlframe.base.functions import asc_nulls_first as asc_nulls_first
from sqlframe.base.functions import asc_nulls_last as asc_nulls_last
from sqlframe.base.functions import ascii as ascii
from sqlframe.base.functions import asin as asin
from sqlframe.base.functions import asinh as asinh
from sqlframe.base.functions import assert_true as assert_true
from sqlframe.base.functions import atan as atan
from sqlframe.base.functions import atan2 as atan2
from sqlframe.base.functions import atanh as atanh
from sqlframe.base.functions import avg as avg
from sqlframe.base.functions import base64 as base64
from sqlframe.base.functions import bin as bin
from sqlframe.base.functions import bit_and as bit_and
from sqlframe.base.functions import bit_count as bit_count
from sqlframe.base.functions import bit_get as bit_get
from sqlframe.base.functions import bit_length as bit_length
from sqlframe.base.functions import bit_or as bit_or
from sqlframe.base.functions import bit_xor as bit_xor
from sqlframe.base.functions import bitmap_bit_position as bitmap_bit_position
from sqlframe.base.functions import bitmap_bucket_number as bitmap_bucket_number
from sqlframe.base.functions import bitmap_construct_agg as bitmap_construct_agg
from sqlframe.base.functions import bitmap_count as bitmap_count
from sqlframe.base.functions import bitmap_or_agg as bitmap_or_agg
from sqlframe.base.functions import bitwise_not as bitwise_not
from sqlframe.base.functions import bitwiseNOT as bitwiseNOT
from sqlframe.base.functions import bool_and as bool_and
from sqlframe.base.functions import bool_or as bool_or
from sqlframe.base.functions import broadcast as broadcast
from sqlframe.base.functions import bround as bround
from sqlframe.base.functions import btrim as btrim
from sqlframe.base.functions import bucket as bucket
from sqlframe.base.functions import call_function as call_function
from sqlframe.base.functions import cardinality as cardinality
from sqlframe.base.functions import cbrt as cbrt
from sqlframe.base.functions import ceil as ceil
from sqlframe.base.functions import ceiling as ceiling
from sqlframe.base.functions import char as char
from sqlframe.base.functions import char_length as char_length
from sqlframe.base.functions import character_length as character_length
from sqlframe.base.functions import coalesce as coalesce
from sqlframe.base.functions import col as col
from sqlframe.base.functions import collect_list as collect_list
from sqlframe.base.functions import collect_set as collect_set
from sqlframe.base.functions import concat as concat
from sqlframe.base.functions import concat_ws as concat_ws
from sqlframe.base.functions import contains as contains
from sqlframe.base.functions import conv as conv
from sqlframe.base.functions import convert_timezone as convert_timezone
from sqlframe.base.functions import corr as corr
from sqlframe.base.functions import cos as cos
from sqlframe.base.functions import cosh as cosh
from sqlframe.base.functions import cot as cot
from sqlframe.base.functions import count as count
from sqlframe.base.functions import count_distinct as count_distinct
from sqlframe.base.functions import count_if as count_if
from sqlframe.base.functions import count_min_sketch as count_min_sketch
from sqlframe.base.functions import countDistinct as countDistinct
from sqlframe.base.functions import covar_pop as covar_pop
from sqlframe.base.functions import covar_samp as covar_samp
from sqlframe.base.functions import crc32 as crc32
from sqlframe.base.functions import create_map as create_map
from sqlframe.base.functions import csc as csc
from sqlframe.base.functions import cume_dist as cume_dist
from sqlframe.base.functions import curdate as curdate
from sqlframe.base.functions import current_catalog as current_catalog
from sqlframe.base.functions import current_database as current_database
from sqlframe.base.functions import current_date as current_date
from sqlframe.base.functions import current_schema as current_schema
from sqlframe.base.functions import current_timestamp as current_timestamp
from sqlframe.base.functions import current_timezone as current_timezone
from sqlframe.base.functions import current_user as current_user
from sqlframe.base.functions import date_add as date_add
from sqlframe.base.functions import date_diff as date_diff
from sqlframe.base.functions import date_format as date_format
from sqlframe.base.functions import date_from_unix_date as date_from_unix_date
from sqlframe.base.functions import date_part as date_part
from sqlframe.base.functions import date_sub as date_sub
from sqlframe.base.functions import date_trunc as date_trunc
from sqlframe.base.functions import dateadd as dateadd
from sqlframe.base.functions import datediff as datediff
from sqlframe.base.functions import datepart as datepart
from sqlframe.base.functions import day as day
from sqlframe.base.functions import dayofmonth as dayofmonth
from sqlframe.base.functions import dayofweek as dayofweek
from sqlframe.base.functions import dayofyear as dayofyear
from sqlframe.base.functions import days as days
from sqlframe.base.functions import decode as decode
from sqlframe.base.functions import degrees as degrees
from sqlframe.base.functions import dense_rank as dense_rank
from sqlframe.base.functions import desc as desc
from sqlframe.base.functions import desc_nulls_first as desc_nulls_first
from sqlframe.base.functions import desc_nulls_last as desc_nulls_last
from sqlframe.base.functions import e as e
from sqlframe.base.functions import element_at as element_at
from sqlframe.base.functions import elt as elt
from sqlframe.base.functions import encode as encode
from sqlframe.base.functions import endswith as endswith
from sqlframe.base.functions import equal_null as equal_null
from sqlframe.base.functions import every as every
from sqlframe.base.functions import exists as exists
from sqlframe.base.functions import exp as exp
from sqlframe.base.functions import explode as explode
from sqlframe.base.functions import explode_outer as explode_outer
from sqlframe.base.functions import expm1 as expm1
from sqlframe.base.functions import expr as expr
from sqlframe.base.functions import extract as extract
from sqlframe.base.functions import factorial as factorial
from sqlframe.base.functions import filter as filter
from sqlframe.base.functions import find_in_set as find_in_set
from sqlframe.base.functions import first as first
from sqlframe.base.functions import first_value as first_value
from sqlframe.base.functions import flatten as flatten
from sqlframe.base.functions import floor as floor
from sqlframe.base.functions import forall as forall
from sqlframe.base.functions import format_number as format_number
from sqlframe.base.functions import format_string as format_string
from sqlframe.base.functions import from_csv as from_csv
from sqlframe.base.functions import from_json as from_json
from sqlframe.base.functions import from_unixtime as from_unixtime
from sqlframe.base.functions import from_utc_timestamp as from_utc_timestamp
from sqlframe.base.functions import get as get
from sqlframe.base.functions import get_active_spark_context as get_active_spark_context
from sqlframe.base.functions import get_json_object as get_json_object
from sqlframe.base.functions import getbit as getbit
from sqlframe.base.functions import greatest as greatest
from sqlframe.base.functions import grouping as grouping
from sqlframe.base.functions import grouping_id as grouping_id
from sqlframe.base.functions import hash as hash
from sqlframe.base.functions import hex as hex
from sqlframe.base.functions import histogram_numeric as histogram_numeric
from sqlframe.base.functions import hll_sketch_agg as hll_sketch_agg
from sqlframe.base.functions import hll_sketch_estimate as hll_sketch_estimate
from sqlframe.base.functions import hll_union as hll_union
from sqlframe.base.functions import hll_union_agg as hll_union_agg
from sqlframe.base.functions import hour as hour
from sqlframe.base.functions import hours as hours
from sqlframe.base.functions import hypot as hypot
from sqlframe.base.functions import ifnull as ifnull
from sqlframe.base.functions import ilike as ilike
from sqlframe.base.functions import initcap as initcap
from sqlframe.base.functions import inline as inline
from sqlframe.base.functions import inline_outer as inline_outer
from sqlframe.base.functions import input_file_name as input_file_name
from sqlframe.base.functions import instr as instr
from sqlframe.base.functions import isnan as isnan
from sqlframe.base.functions import isnotnull as isnotnull
from sqlframe.base.functions import isnull as isnull
from sqlframe.base.functions import java_method as java_method
from sqlframe.base.functions import json_array_length as json_array_length
from sqlframe.base.functions import json_object_keys as json_object_keys
from sqlframe.base.functions import json_tuple as json_tuple
from sqlframe.base.functions import kurtosis as kurtosis
from sqlframe.base.functions import lag as lag
from sqlframe.base.functions import last as last
from sqlframe.base.functions import last_day as last_day
from sqlframe.base.functions import last_value as last_value
from sqlframe.base.functions import lcase as lcase
from sqlframe.base.functions import lead as lead
from sqlframe.base.functions import least as least
from sqlframe.base.functions import left as left
from sqlframe.base.functions import length as length
from sqlframe.base.functions import levenshtein as levenshtein
from sqlframe.base.functions import like as like
from sqlframe.base.functions import lit as lit
from sqlframe.base.functions import ln as ln
from sqlframe.base.functions import localtimestamp as localtimestamp
from sqlframe.base.functions import locate as locate
from sqlframe.base.functions import log as log
from sqlframe.base.functions import log1p as log1p
from sqlframe.base.functions import log2 as log2
from sqlframe.base.functions import log10 as log10
from sqlframe.base.functions import lower as lower
from sqlframe.base.functions import lpad as lpad
from sqlframe.base.functions import ltrim as ltrim
from sqlframe.base.functions import make_date as make_date
from sqlframe.base.functions import make_dt_interval as make_dt_interval
from sqlframe.base.functions import make_interval as make_interval
from sqlframe.base.functions import make_timestamp as make_timestamp
from sqlframe.base.functions import make_timestamp_ltz as make_timestamp_ltz
from sqlframe.base.functions import make_timestamp_ntz as make_timestamp_ntz
from sqlframe.base.functions import make_ym_interval as make_ym_interval
from sqlframe.base.functions import map_concat as map_concat
from sqlframe.base.functions import map_contains_key as map_contains_key
from sqlframe.base.functions import map_entries as map_entries
from sqlframe.base.functions import map_filter as map_filter
from sqlframe.base.functions import map_from_arrays as map_from_arrays
from sqlframe.base.functions import map_from_entries as map_from_entries
from sqlframe.base.functions import map_keys as map_keys
from sqlframe.base.functions import map_values as map_values
from sqlframe.base.functions import map_zip_with as map_zip_with
from sqlframe.base.functions import mask as mask
from sqlframe.base.functions import max as max
from sqlframe.base.functions import max_by as max_by
from sqlframe.base.functions import md5 as md5
from sqlframe.base.functions import mean as mean
from sqlframe.base.functions import median as median
from sqlframe.base.functions import min as min
from sqlframe.base.functions import min_by as min_by
from sqlframe.base.functions import minute as minute
from sqlframe.base.functions import mode as mode
from sqlframe.base.functions import (
    monotonically_increasing_id as monotonically_increasing_id,
)
from sqlframe.base.functions import month as month
from sqlframe.base.functions import months as months
from sqlframe.base.functions import months_between as months_between
from sqlframe.base.functions import named_struct as named_struct
from sqlframe.base.functions import nanvl as nanvl
from sqlframe.base.functions import negate as negate
from sqlframe.base.functions import negative as negative
from sqlframe.base.functions import next_day as next_day
from sqlframe.base.functions import now as now
from sqlframe.base.functions import nth_value as nth_value
from sqlframe.base.functions import ntile as ntile
from sqlframe.base.functions import nullif as nullif
from sqlframe.base.functions import nvl as nvl
from sqlframe.base.functions import nvl2 as nvl2
from sqlframe.base.functions import octet_length as octet_length
from sqlframe.base.functions import overlay as overlay
from sqlframe.base.functions import parse_url as parse_url
from sqlframe.base.functions import percent_rank as percent_rank
from sqlframe.base.functions import percentile as percentile
from sqlframe.base.functions import percentile_approx as percentile_approx
from sqlframe.base.functions import pi as pi
from sqlframe.base.functions import pmod as pmod
from sqlframe.base.functions import posexplode as posexplode
from sqlframe.base.functions import posexplode_outer as posexplode_outer
from sqlframe.base.functions import position as position
from sqlframe.base.functions import positive as positive
from sqlframe.base.functions import pow as pow
from sqlframe.base.functions import power as power
from sqlframe.base.functions import printf as printf
from sqlframe.base.functions import quarter as quarter
from sqlframe.base.functions import radians as radians
from sqlframe.base.functions import raise_error as raise_error
from sqlframe.base.functions import rand as rand
from sqlframe.base.functions import randn as randn
from sqlframe.base.functions import rank as rank
from sqlframe.base.functions import reduce as reduce
from sqlframe.base.functions import reflect as reflect
from sqlframe.base.functions import regexp as regexp
from sqlframe.base.functions import regexp_count as regexp_count
from sqlframe.base.functions import regexp_extract as regexp_extract
from sqlframe.base.functions import regexp_extract_all as regexp_extract_all
from sqlframe.base.functions import regexp_instr as regexp_instr
from sqlframe.base.functions import regexp_like as regexp_like
from sqlframe.base.functions import regexp_replace as regexp_replace
from sqlframe.base.functions import regexp_substr as regexp_substr
from sqlframe.base.functions import regr_avgx as regr_avgx
from sqlframe.base.functions import regr_avgy as regr_avgy
from sqlframe.base.functions import regr_count as regr_count
from sqlframe.base.functions import regr_intercept as regr_intercept
from sqlframe.base.functions import regr_r2 as regr_r2
from sqlframe.base.functions import regr_slope as regr_slope
from sqlframe.base.functions import regr_sxx as regr_sxx
from sqlframe.base.functions import regr_sxy as regr_sxy
from sqlframe.base.functions import regr_syy as regr_syy
from sqlframe.base.functions import repeat as repeat
from sqlframe.base.functions import replace as replace
from sqlframe.base.functions import reverse as reverse
from sqlframe.base.functions import right as right
from sqlframe.base.functions import rint as rint
from sqlframe.base.functions import rlike as rlike
from sqlframe.base.functions import round as round
from sqlframe.base.functions import row_number as row_number
from sqlframe.base.functions import rpad as rpad
from sqlframe.base.functions import rtrim as rtrim
from sqlframe.base.functions import schema_of_csv as schema_of_csv
from sqlframe.base.functions import schema_of_json as schema_of_json
from sqlframe.base.functions import sec as sec
from sqlframe.base.functions import second as second
from sqlframe.base.functions import sentences as sentences
from sqlframe.base.functions import sequence as sequence
from sqlframe.base.functions import sha as sha
from sqlframe.base.functions import sha1 as sha1
from sqlframe.base.functions import sha2 as sha2
from sqlframe.base.functions import shiftLeft as shiftLeft
from sqlframe.base.functions import shiftleft as shiftleft
from sqlframe.base.functions import shiftRight as shiftRight
from sqlframe.base.functions import shiftright as shiftright
from sqlframe.base.functions import shiftRightUnsigned as shiftRightUnsigned
from sqlframe.base.functions import shiftrightunsigned as shiftrightunsigned
from sqlframe.base.functions import shuffle as shuffle
from sqlframe.base.functions import sign as sign
from sqlframe.base.functions import signum as signum
from sqlframe.base.functions import sin as sin
from sqlframe.base.functions import sinh as sinh
from sqlframe.base.functions import size as size
from sqlframe.base.functions import skewness as skewness
from sqlframe.base.functions import slice as slice
from sqlframe.base.functions import some as some
from sqlframe.base.functions import sort_array as sort_array
from sqlframe.base.functions import soundex as soundex
from sqlframe.base.functions import spark_partition_id as spark_partition_id
from sqlframe.base.functions import split as split
from sqlframe.base.functions import split_part as split_part
from sqlframe.base.functions import sqrt as sqrt
from sqlframe.base.functions import stack as stack
from sqlframe.base.functions import startswith as startswith
from sqlframe.base.functions import std as std
from sqlframe.base.functions import stddev as stddev
from sqlframe.base.functions import stddev_pop as stddev_pop
from sqlframe.base.functions import stddev_samp as stddev_samp
from sqlframe.base.functions import str_to_map as str_to_map
from sqlframe.base.functions import struct as struct
from sqlframe.base.functions import substr as substr
from sqlframe.base.functions import substring as substring
from sqlframe.base.functions import substring_index as substring_index
from sqlframe.base.functions import sum as sum
from sqlframe.base.functions import sum_distinct as sum_distinct
from sqlframe.base.functions import sumDistinct as sumDistinct
from sqlframe.base.functions import tan as tan
from sqlframe.base.functions import tanh as tanh
from sqlframe.base.functions import timestamp_micros as timestamp_micros
from sqlframe.base.functions import timestamp_millis as timestamp_millis
from sqlframe.base.functions import timestamp_seconds as timestamp_seconds
from sqlframe.base.functions import to_binary as to_binary
from sqlframe.base.functions import to_char as to_char
from sqlframe.base.functions import to_csv as to_csv
from sqlframe.base.functions import to_date as to_date
from sqlframe.base.functions import to_json as to_json
from sqlframe.base.functions import to_number as to_number
from sqlframe.base.functions import to_timestamp as to_timestamp
from sqlframe.base.functions import to_timestamp_ltz as to_timestamp_ltz
from sqlframe.base.functions import to_timestamp_ntz as to_timestamp_ntz
from sqlframe.base.functions import to_unix_timestamp as to_unix_timestamp
from sqlframe.base.functions import to_utc_timestamp as to_utc_timestamp
from sqlframe.base.functions import to_varchar as to_varchar
from sqlframe.base.functions import toDegrees as toDegrees
from sqlframe.base.functions import toRadians as toRadians
from sqlframe.base.functions import transform as transform
from sqlframe.base.functions import transform_keys as transform_keys
from sqlframe.base.functions import transform_values as transform_values
from sqlframe.base.functions import translate as translate
from sqlframe.base.functions import trim as trim
from sqlframe.base.functions import trunc as trunc
from sqlframe.base.functions import try_add as try_add
from sqlframe.base.functions import try_aes_decrypt as try_aes_decrypt
from sqlframe.base.functions import try_avg as try_avg
from sqlframe.base.functions import try_divide as try_divide
from sqlframe.base.functions import try_element_at as try_element_at
from sqlframe.base.functions import try_multiply as try_multiply
from sqlframe.base.functions import try_subtract as try_subtract
from sqlframe.base.functions import try_sum as try_sum
from sqlframe.base.functions import try_to_binary as try_to_binary
from sqlframe.base.functions import try_to_number as try_to_number
from sqlframe.base.functions import try_to_timestamp as try_to_timestamp
from sqlframe.base.functions import typeof as typeof
from sqlframe.base.functions import ucase as ucase
from sqlframe.base.functions import unbase64 as unbase64
from sqlframe.base.functions import unhex as unhex
from sqlframe.base.functions import unix_date as unix_date
from sqlframe.base.functions import unix_micros as unix_micros
from sqlframe.base.functions import unix_millis as unix_millis
from sqlframe.base.functions import unix_seconds as unix_seconds
from sqlframe.base.functions import unix_timestamp as unix_timestamp
from sqlframe.base.functions import upper as upper
from sqlframe.base.functions import url_decode as url_decode
from sqlframe.base.functions import url_encode as url_encode
from sqlframe.base.functions import user as user
from sqlframe.base.functions import var_pop as var_pop
from sqlframe.base.functions import var_samp as var_samp
from sqlframe.base.functions import variance as variance
from sqlframe.base.functions import version as version
from sqlframe.base.functions import weekday as weekday
from sqlframe.base.functions import weekofyear as weekofyear
from sqlframe.base.functions import when as when
from sqlframe.base.functions import width_bucket as width_bucket
from sqlframe.base.functions import xpath as xpath
from sqlframe.base.functions import xpath_boolean as xpath_boolean
from sqlframe.base.functions import xpath_double as xpath_double
from sqlframe.base.functions import xpath_float as xpath_float
from sqlframe.base.functions import xpath_int as xpath_int
from sqlframe.base.functions import xpath_long as xpath_long
from sqlframe.base.functions import xpath_number as xpath_number
from sqlframe.base.functions import xpath_short as xpath_short
from sqlframe.base.functions import xpath_string as xpath_string
from sqlframe.base.functions import xxhash64 as xxhash64
from sqlframe.base.functions import year as year
from sqlframe.base.functions import years as years
from sqlframe.base.functions import zip_with as zip_with



from sqlframe.base.table import _BaseTable
from sqlframe.standalone.dataframe import StandaloneDataFrame


class StandaloneTable(StandaloneDataFrame, _BaseTable["StandaloneDataFrame"]):
    _df = StandaloneDataFrame



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlframe.base.mixins.readwriter_mixins import PandasLoaderMixin, PandasWriterMixin
from sqlframe.base.readerwriter import (
    _BaseDataFrameReader,
    _BaseDataFrameWriter,
)

if t.TYPE_CHECKING:
    from sqlframe.spark.dataframe import SparkDataFrame
    from sqlframe.spark.session import SparkSession
    from sqlframe.spark.table import SparkTable


class SparkDataFrameReader(
    PandasLoaderMixin["SparkSession", "SparkDataFrame"],
    _BaseDataFrameReader["SparkSession", "SparkDataFrame", "SparkTable"],
):
    pass


class SparkDataFrameWriter(
    PandasWriterMixin["SparkSession", "SparkDataFrame"],
    _BaseDataFrameWriter["SparkSession", "SparkDataFrame"],
):
    pass



import functools
import logging
import typing as t

from sqlglot import exp

try:
    from sqlglot.expressions import Whens
except ImportError:
    Whens = None  # type: ignore
from sqlglot.helper import object_to_dict

from sqlframe.base.column import Column
from sqlframe.base.table import (
    DF,
    Clause,
    LazyExpression,
    WhenMatched,
    WhenNotMatched,
    WhenNotMatchedBySource,
    _BaseTable,
)

if t.TYPE_CHECKING:
    from sqlframe.base._typing import ColumnOrLiteral


logger = logging.getLogger(__name__)


def ensure_cte() -> t.Callable[[t.Callable], t.Callable]:
    def decorator(func: t.Callable) -> t.Callable:
        @functools.wraps(func)
        def wrapper(self: _BaseTable, *args, **kwargs) -> t.Any:
            if len(self.expression.ctes) > 0:
                return func(self, *args, **kwargs)  # type: ignore
            self_class = self.__class__
            self = self._convert_leaf_to_cte()
            self = self_class(**object_to_dict(self))
            return func(self, *args, **kwargs)  # type: ignore

        wrapper.__wrapped__ = func  # type: ignore
        return wrapper

    return decorator


class _BaseTableMixins(_BaseTable, t.Generic[DF]):
    def _ensure_where_condition(
        self, where: t.Optional[t.Union[Column, str, bool]] = None
    ) -> exp.Expression:
        self_name = self.expression.ctes[0].this.args["from"].this.alias_or_name

        if where is None:
            logger.warning("Empty value for `where`clause. Defaults to `True`.")
            condition: exp.Expression = exp.Boolean(this=True)
        else:
            condition_list = self._ensure_and_normalize_cols(where, self.expression)
            if len(condition_list) > 1:
                condition_list = [functools.reduce(lambda x, y: x & y, condition_list)]
            for col_expr in condition_list[0].expression.find_all(exp.Column):
                if col_expr.table == self.expression.args["from"].this.alias_or_name:
                    col_expr.set("table", exp.to_identifier(self_name))
            condition = condition_list[0].expression
            if isinstance(condition, exp.Alias):
                condition = condition.this
        return condition


class UpdateSupportMixin(_BaseTableMixins, t.Generic[DF]):
    @ensure_cte()
    def update(
        self,
        set_: t.Dict[t.Union[Column, str], t.Union[Column, "ColumnOrLiteral", exp.Expression]],
        where: t.Optional[t.Union[Column, str, bool]] = None,
    ) -> LazyExpression:
        self_expr = self.expression.ctes[0].this.args["from"].this

        condition = self._ensure_where_condition(where)
        update_set = self._ensure_and_normalize_update_set(set_)
        update_expr = exp.Update(
            this=self_expr,
            expressions=[
                exp.EQ(
                    this=key,
                    expression=val,
                )
                for key, val in update_set.items()
            ],
            where=exp.Where(this=condition),
        )

        return LazyExpression(update_expr, self.session)

    def _ensure_and_normalize_update_set(
        self,
        set_: t.Dict[t.Union[Column, str], t.Union[Column, "ColumnOrLiteral", exp.Expression]],
    ) -> t.Dict[str, exp.Expression]:
        self_name = self.expression.ctes[0].this.args["from"].this.alias_or_name
        update_set = {}
        for key, val in set_.items():
            key_column: Column = self._ensure_and_normalize_col(key)
            key_expr = list(key_column.expression.find_all(exp.Column))
            if len(key_expr) > 1:
                raise ValueError(f"Can only update one a single column at a time.")
            key = key_expr[0].alias_or_name

            val_column: Column = self._ensure_and_normalize_col(val)
            for col_expr in val_column.expression.find_all(exp.Column):
                if col_expr.table == self.expression.args["from"].this.alias_or_name:
                    col_expr.set("table", exp.to_identifier(self_name))
                else:
                    raise ValueError(
                        f"Column `{col_expr.alias_or_name}` does not exist in the table."
                    )

            update_set[key] = val_column.expression
        return update_set


class DeleteSupportMixin(_BaseTableMixins, t.Generic[DF]):
    @ensure_cte()
    def delete(
        self,
        where: t.Optional[t.Union[Column, str, bool]] = None,
    ) -> LazyExpression:
        self_expr = self.expression.ctes[0].this.args["from"].this

        condition = self._ensure_where_condition(where)
        delete_expr = exp.Delete(
            this=self_expr,
            where=exp.Where(this=condition),
        )

        return LazyExpression(delete_expr, self.session)


class MergeSupportMixin(_BaseTable, t.Generic[DF]):
    _merge_supported_clauses: t.Iterable[
        t.Union[t.Type[WhenMatched], t.Type[WhenNotMatched], t.Type[WhenNotMatchedBySource]]
    ]
    _merge_support_star: bool

    @ensure_cte()
    def merge(
        self,
        other_df: DF,
        condition: t.Union[str, t.List[str], Column, t.List[Column], bool],
        clauses: t.Iterable[t.Union[WhenMatched, WhenNotMatched, WhenNotMatchedBySource]],
    ) -> LazyExpression:
        self_name = self.expression.ctes[0].this.args["from"].this.alias_or_name
        self_expr = self.expression.ctes[0].this.args["from"].this

        other_df = other_df._convert_leaf_to_cte()

        if condition is None:
            raise ValueError("condition cannot be None")

        condition_columns: Column = self._ensure_and_normalize_condition(condition, other_df)
        other_name = self._create_hash_from_expression(other_df.expression)
        other_expr = exp.Subquery(
            this=other_df.expression, alias=exp.TableAlias(this=exp.to_identifier(other_name))
        )

        for col_expr in condition_columns.expression.find_all(exp.Column):
            if col_expr.table == self.expression.args["from"].this.alias_or_name:
                col_expr.set("table", exp.to_identifier(self_name))
            if col_expr.table == other_df.latest_cte_name:
                col_expr.set("table", exp.to_identifier(other_name))

        merge_expressions = []
        for clause in clauses:
            if not isinstance(clause, tuple(self._merge_supported_clauses)):
                raise ValueError(
                    f"Unsupported clause type {type(clause.clause)} for merge operation"
                )
            expression = None

            if clause.clause.condition is not None:
                cond_clause = self._ensure_and_normalize_condition(
                    clause.clause.condition, other_df, True
                )
                for col_expr in cond_clause.expression.find_all(exp.Column):
                    if col_expr.table == self.expression.args["from"].this.alias_or_name:
                        col_expr.set("table", exp.to_identifier(self_name))
                    if col_expr.table == other_df.latest_cte_name:
                        col_expr.set("table", exp.to_identifier(other_name))
            else:
                cond_clause = None
            if clause.clause.clause_type == Clause.UPDATE:
                update_set = self._ensure_and_normalize_assignments(
                    clause.clause.assignments, other_df
                )
                expression = exp.When(
                    matched=clause.clause.matched,
                    source=clause.clause.by_source,
                    condition=cond_clause.expression if cond_clause else None,
                    then=exp.Update(
                        expressions=[
                            exp.EQ(
                                this=key,
                                expression=val,
                            )
                            for key, val in update_set.items()
                        ]
                    ),
                )
            if clause.clause.clause_type == Clause.UPDATE_ALL:
                if not self._support_star:
                    raise ValueError("Merge operation does not support UPDATE_ALL")
                expression = exp.When(
                    matched=clause.clause.matched,
                    source=clause.clause.by_source,
                    condition=cond_clause.expression if cond_clause else None,
                    then=exp.Update(expressions=[exp.Star()]),
                )
            elif clause.clause.clause_type == Clause.INSERT:
                insert_values = self._ensure_and_normalize_assignments(
                    clause.clause.assignments, other_df
                )
                expression = exp.When(
                    matched=clause.clause.matched,
                    source=clause.clause.by_source,
                    condition=cond_clause.expression if cond_clause else None,
                    then=exp.Insert(
                        this=exp.Tuple(expressions=[key for key in insert_values.keys()]),
                        expression=exp.Tuple(expressions=[val for val in insert_values.values()]),
                    ),
                )
            elif clause.clause.clause_type == Clause.INSERT_ALL:
                if not self._support_star:
                    raise ValueError("Merge operation does not support INSERT_ALL")
                expression = exp.When(
                    matched=clause.clause.matched,
                    source=clause.clause.by_source,
                    condition=cond_clause.expression if cond_clause else None,
                    then=exp.Insert(expression=exp.Star()),
                )
            elif clause.clause.clause_type == Clause.DELETE:
                expression = exp.When(
                    matched=clause.clause.matched,
                    source=clause.clause.by_source,
                    condition=cond_clause.expression if cond_clause else None,
                    then=exp.var("DELETE"),
                )

            if expression:
                merge_expressions.append(expression)

        if Whens is None:
            merge_expr = exp.merge(
                *merge_expressions,
                into=self_expr,
                using=other_expr,
                on=condition_columns.expression,
            )
        else:
            merge_expr = exp.merge(
                Whens(expressions=merge_expressions),
                into=self_expr,
                using=other_expr,
                on=condition_columns.expression,
            )

        return LazyExpression(merge_expr, self.session)

    def _ensure_and_normalize_condition(
        self,
        condition: t.Union[str, t.List[str], Column, t.List[Column], bool],
        other_df: DF,
        clause: t.Optional[bool] = False,
    ):
        join_expression = self._add_ctes_to_expression(
            self.expression, other_df.expression.copy().ctes
        )
        condition = self._ensure_and_normalize_cols(condition, self.expression)
        self._handle_self_join(other_df, condition)

        if isinstance(condition[0].expression, exp.Column) and not clause:
            table_names = [
                table.alias_or_name
                for table in [
                    self.expression.args["from"].this,
                    other_df.expression.args["from"].this,
                ]
            ]

            join_column_pairs, join_clause = self._handle_join_column_names_only(
                condition, join_expression, other_df, table_names
            )
        else:
            join_clause = self._normalize_join_clause(condition, join_expression)
        return join_clause

    def _ensure_and_normalize_assignments(
        self,
        assignments: t.Dict[
            t.Union[Column, str], t.Union[Column, "ColumnOrLiteral", exp.Expression]
        ],
        other_df,
    ) -> t.Dict[exp.Column, exp.Expression]:
        self_name = self.expression.ctes[0].this.args["from"].this.alias_or_name
        other_name = self._create_hash_from_expression(other_df.expression)
        update_set = {}
        for key, val in assignments.items():
            key_column: Column = self._ensure_and_normalize_col(key)
            key_expr = list(key_column.expression.find_all(exp.Column))
            if len(key_expr) > 1:
                raise ValueError(f"Target expression `{key_expr}` should be a single column.")
            column_key = exp.column(key_expr[0].alias_or_name)

            val = self._ensure_and_normalize_col(val)
            val = self._ensure_and_normalize_cols(val, other_df.expression)[0]
            if self.branch_id == other_df.branch_id:
                other_df_unique_uuids = other_df.known_uuids - self.known_uuids
                for col_expr in val.expression.find_all(exp.Column):
                    if (
                        "join_on_uuid" in col_expr.meta
                        and col_expr.meta["join_on_uuid"] in other_df_unique_uuids
                    ):
                        col_expr.set("table", exp.to_identifier(other_df.latest_cte_name))

            for col_expr in val.expression.find_all(exp.Column):
                if not col_expr.table or col_expr.table == other_df.latest_cte_name:
                    col_expr.set("table", exp.to_identifier(other_name))
                elif col_expr.table == self.expression.args["from"].this.alias_or_name:
                    col_expr.set("table", exp.to_identifier(self_name))
                else:
                    raise ValueError(
                        f"Column `{col_expr.alias_or_name}` does not exist in any of the tables."
                    )
            if isinstance(val.expression, exp.Alias):
                val.expression = val.expression.this
            update_set[column_key] = val.expression
        return update_set



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlframe.base.mixins.readwriter_mixins import PandasLoaderMixin, PandasWriterMixin
from sqlframe.base.readerwriter import (
    _BaseDataFrameReader,
    _BaseDataFrameWriter,
)

if t.TYPE_CHECKING:
    from sqlframe.bigquery.session import BigQuerySession  # noqa
    from sqlframe.bigquery.dataframe import BigQueryDataFrame  # noqa
    from sqlframe.bigquery.table import BigQueryTable  # noqa


class BigQueryDataFrameReader(
    PandasLoaderMixin["BigQuerySession", "BigQueryDataFrame"],
    _BaseDataFrameReader["BigQuerySession", "BigQueryDataFrame", "BigQueryTable"],
):
    pass


class BigQueryDataFrameWriter(
    PandasWriterMixin["BigQuerySession", "BigQueryDataFrame"],
    _BaseDataFrameWriter["BigQuerySession", "BigQueryDataFrame"],
):
    pass



import inspect
import sys

import sqlframe.base.functions  # noqa

module = sys.modules["sqlframe.base.functions"]
globals().update(
    {
        name: func
        for name, func in inspect.getmembers(module, inspect.isfunction)
        if hasattr(func, "unsupported_engines") and "spark" not in func.unsupported_engines
    }
)



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlframe.base.mixins.readwriter_mixins import PandasLoaderMixin, PandasWriterMixin
from sqlframe.base.readerwriter import (
    _BaseDataFrameReader,
    _BaseDataFrameWriter,
)

if t.TYPE_CHECKING:
    from sqlframe.snowflake.session import SnowflakeSession  # noqa
    from sqlframe.snowflake.dataframe import SnowflakeDataFrame  # noqa
    from sqlframe.snowflake.table import SnowflakeTable  # noqa


class SnowflakeDataFrameReader(
    PandasLoaderMixin["SnowflakeSession", "SnowflakeDataFrame"],
    _BaseDataFrameReader["SnowflakeSession", "SnowflakeDataFrame", "SnowflakeTable"],
):
    pass


class SnowflakeDataFrameWriter(
    PandasWriterMixin["SnowflakeSession", "SnowflakeDataFrame"],
    _BaseDataFrameWriter["SnowflakeSession", "SnowflakeDataFrame"],
):
    pass



import typing as t

import pytest

from sqlframe.base.catalog import CatalogMetadata, Column, Database, Function, Table
from sqlframe.bigquery.session import BigQuerySession

pytest_plugins = ["tests.integration.fixtures"]
pytestmark = [
    pytest.mark.bigquery,
    pytest.mark.xdist_group("bigquery_tests"),
]


@pytest.fixture
def reset_catalog(bigquery_session: BigQuerySession) -> t.Iterator[None]:
    yield
    bigquery_session.catalog.setCurrentCatalog("sqlframe")


@pytest.fixture
def reset_database(bigquery_session: BigQuerySession) -> t.Iterator[None]:
    yield
    bigquery_session.catalog.setCurrentDatabase("sqlframe.db1")


def test_current_catalog(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.currentCatalog() == "sqlframe"


def test_set_current_catalog(bigquery_session: BigQuerySession, reset_catalog: t.Iterator[None]):
    assert bigquery_session.catalog.currentCatalog() == "sqlframe"
    bigquery_session.catalog.setCurrentCatalog("catalog1")
    assert bigquery_session.catalog.currentCatalog() == "catalog1"


def test_list_catalogs(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.listCatalogs() == [
        CatalogMetadata(name="sqlframe", description=None)
    ]


def test_current_database(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.currentDatabase() == "db1"


def test_set_current_database(bigquery_session: BigQuerySession, reset_database: t.Iterator[None]):
    assert bigquery_session.catalog.currentDatabase() == "db1"
    bigquery_session.catalog.setCurrentDatabase("sqlframe.db2")
    assert bigquery_session.catalog.currentDatabase() == "db2"


def test_list_databases(bigquery_session: BigQuerySession):
    assert sorted(bigquery_session.catalog.listDatabases(), key=lambda x: (x.catalog, x.name)) == [
        Database(name="db1", catalog="sqlframe", description=None, locationUri=""),
    ]


def test_list_databases_pattern(bigquery_session: BigQuerySession):
    assert sorted(
        bigquery_session.catalog.listDatabases("db*"), key=lambda x: (x.catalog, x.name)
    ) == [
        Database(name="db1", catalog="sqlframe", description=None, locationUri=""),
    ]


def test_get_database_no_match(bigquery_session: BigQuerySession):
    with pytest.raises(ValueError):
        assert bigquery_session.catalog.getDatabase("nonexistent")


def test_get_database_name_only(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.getDatabase("db1") == Database(
        name="db1", catalog="sqlframe", description=None, locationUri=""
    )


def test_get_database_name_and_catalog(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.getDatabase("db1") == Database(
        name="db1", catalog="sqlframe", description=None, locationUri=""
    )


def test_database_exists_does_exist(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.databaseExists("db1") is True


def test_database_exists_does_not_exist(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.databaseExists("nonexistent") is False


def test_list_tables_no_args(bigquery_session: BigQuerySession):
    assert sorted(
        bigquery_session.catalog.listTables(), key=lambda x: (x.catalog, x.database, x.name)
    ) == [
        Table(
            name="table1",
            catalog="sqlframe",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_no_catalog(bigquery_session: BigQuerySession):
    assert sorted(
        bigquery_session.catalog.listTables("db1"), key=lambda x: (x.catalog, x.database, x.name)
    ) == [
        Table(
            name="table1",
            catalog="sqlframe",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_and_catalog(bigquery_session: BigQuerySession):
    assert sorted(
        bigquery_session.catalog.listTables("sqlframe.db1"),
        key=lambda x: (x.catalog, x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="sqlframe",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_pattern(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.listTables(pattern="tab*") == [
        Table(
            name="table1",
            catalog="sqlframe",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_get_table(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.getTable("table1") == Table(
        name="table1",
        catalog="sqlframe",
        namespace=["db1"],
        description=None,
        tableType="MANAGED",
        isTemporary=False,
    )


def test_get_table_not_exists(bigquery_session: BigQuerySession):
    with pytest.raises(ValueError):
        assert bigquery_session.catalog.getTable("nonexistent")


def test_list_functions(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.listFunctions() == [
        Function(
            name="add",
            catalog="sqlframe",
            namespace=["db1"],
            description=None,
            className="",
            isTemporary=False,
        )
    ]


def test_list_functions_pattern(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.listFunctions(pattern="a*") == [
        Function(
            name="add",
            catalog="sqlframe",
            namespace=["db1"],
            description=None,
            className="",
            isTemporary=False,
        )
    ]


def test_list_functions_db_name_and_pattern(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.listFunctions("db1", pattern="a*") == [
        Function(
            name="add",
            catalog="sqlframe",
            namespace=["db1"],
            description=None,
            className="",
            isTemporary=False,
        )
    ]


def test_function_exists_does_exist(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.functionExists("add") is True


def test_function_exists_does_not_exist(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.functionExists("nonexistent") is False


def test_get_function_exists(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.getFunction("sqlframe.db1.add") == Function(
        name="add",
        catalog="sqlframe",
        namespace=["db1"],
        description=None,
        className="",
        isTemporary=False,
    )


def test_get_function_not_exists(bigquery_session: BigQuerySession):
    with pytest.raises(ValueError):
        assert bigquery_session.catalog.getFunction("sqlframe.db1.nonexistent")


def test_list_columns(bigquery_session: BigQuerySession):
    assert sorted(bigquery_session.catalog.listColumns("table1"), key=lambda x: x.name) == [
        Column(
            name="id",
            description=None,
            dataType="BIGINT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_list_columns_use_db_name(bigquery_session: BigQuerySession):
    assert sorted(
        bigquery_session.catalog.listColumns("table1", dbName="sqlframe.db1"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="BIGINT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_table_exists_table_name_only(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.tableExists("sqlframe.db1.table1") is True


def test_table_exists_table_name_and_db_name(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.tableExists("table1", dbName="sqlframe.db1") is True


def test_table_not_exists(bigquery_session: BigQuerySession):
    assert bigquery_session.catalog.tableExists("nonexistent", dbName="sqlframe.db1") is False


def test_create_external_table(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.createExternalTable("table1", "sqlframe.db1", "path/to/table")


def test_create_table(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.createTable("table1", "sqlframe.db1")


def test_drop_temp_view(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.dropTempView("view1")


def test_drop_global_temp_view(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.dropGlobalTempView("view1")


def test_register_function(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.registerFunction("function1", lambda x: x)


def test_is_cached(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.isCached("table1")


def test_cache_table(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.cacheTable("table1")


def test_uncache_table(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.uncacheTable("table1")


def test_clear_cache(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.clearCache()


def test_refresh_table(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.refreshTable("table1")


def test_recover_partitions(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.recoverPartitions("table1")


def test_refresh_by_path(bigquery_session: BigQuerySession):
    with pytest.raises(NotImplementedError):
        bigquery_session.catalog.refreshByPath("path/to/table")



from sqlframe.base.mixins.table_mixins import (
    DeleteSupportMixin,
    UpdateSupportMixin,
)
from sqlframe.base.table import _BaseTable
from sqlframe.redshift.dataframe import RedshiftDataFrame


class RedshiftTable(
    RedshiftDataFrame,
    UpdateSupportMixin["RedshiftDataFrame"],
    DeleteSupportMixin["RedshiftDataFrame"],
    _BaseTable["RedshiftDataFrame"],
):
    _df = RedshiftDataFrame



from sqlframe.base.mixins.table_mixins import (
    DeleteSupportMixin,
    MergeSupportMixin,
    UpdateSupportMixin,
)
from sqlframe.base.table import _BaseTable
from sqlframe.duckdb.dataframe import DuckDBDataFrame


class DuckDBTable(
    DuckDBDataFrame,
    UpdateSupportMixin["DuckDBDataFrame"],
    DeleteSupportMixin["DuckDBDataFrame"],
    _BaseTable["DuckDBDataFrame"],
):
    _df = DuckDBDataFrame



from __future__ import annotations

import datetime
import typing as t

import pytest
from pyspark.sql import SparkSession as PySparkSession

from sqlframe.base.table import WhenMatched, WhenNotMatched, _BaseTable
from sqlframe.base.types import Row
from sqlframe.bigquery.session import BigQuerySession
from sqlframe.databricks import DatabricksSession
from sqlframe.duckdb.session import DuckDBSession
from sqlframe.postgres.session import PostgresSession
from sqlframe.redshift.session import RedshiftSession
from sqlframe.snowflake.session import SnowflakeSession
from sqlframe.spark.session import SparkSession
from sqlframe.standalone.session import StandaloneSession

if t.TYPE_CHECKING:
    from sqlframe.base.dataframe import BaseDataFrame

pytest_plugins = ["tests.integration.fixtures"]


@pytest.fixture
def merge_data() -> t.Tuple[t.List[t.Any], str]:
    return (
        [
            (
                "3368c22d-edd8-4ae9-a0b8-f0956b9ffa88",
                1,
                "Jack",
                "Shephard",
                37,
                1,
                "1999-01-01",
                "2900-01-01",
            ),
            (
                "6db4c8ce-464d-4772-a686-e6bd9e10286f",
                2,
                "John",
                "Locke",
                65,
                1,
                "1999-01-01",
                "2900-01-01",
            ),
            (
                "7984b6b1-cb4c-49a0-bc83-61064f4d77e6",
                3,
                "Kate",
                "Austen",
                37,
                2,
                "1999-01-01",
                "2900-01-01",
            ),
            (
                "8ee4d1bf-c3c7-42a0-8727-233881d6212e",
                4,
                "Claire",
                "Littleton",
                27,
                2,
                "1999-01-01",
                "2900-01-01",
            ),
            (
                "0c67fff1-9471-4360-8624-982b009cf315",
                5,
                "Hugo",
                "Reyes",
                29,
                100,
                "1999-01-01",
                "2900-01-01",
            ),
        ],
        "s_id STRING, employee_id INTEGER, fname STRING, lname STRING, age INTEGER, "
        "store_id INTEGER, start_date DATE, end_date DATE",
    )


@pytest.fixture
def cleanup_employee_df(
    get_engine_df: t.Callable[[str], BaseDataFrame],
) -> t.Iterator[BaseDataFrame]:
    df = get_engine_df("employee")
    df.session._execute("DROP TABLE IF EXISTS update_employee")
    df.session._execute("DROP TABLE IF EXISTS merge_employee")
    df.session._execute("DROP TABLE IF EXISTS delete_employee")
    yield df
    df.session._execute("DROP TABLE IF EXISTS update_employee")
    df.session._execute("DROP TABLE IF EXISTS merge_employee")
    df.session._execute("DROP TABLE IF EXISTS delete_employee")


def test_update_table(cleanup_employee_df: BaseDataFrame, caplog):
    session = cleanup_employee_df.session
    if isinstance(
        session,
        (
            StandaloneSession,
            PySparkSession,
            SparkSession,
        ),
    ):
        pytest.skip("Engine doesn't support update")
    df_employee = cleanup_employee_df
    df_employee.write.saveAsTable("update_employee")
    df = session.read.table("update_employee")
    assert isinstance(df, _BaseTable)
    update_expr = df.update(
        set_={"age": df["age"] + 1},
        where=df["employee_id"] == 1,
    )
    result = update_expr.execute()
    # Postgres, RedshiftSession and BigQuery don't support returning the number of affected rows
    if not isinstance(session, (PostgresSession, BigQuerySession, RedshiftSession)):
        assert result[0][0] == 1

    df2 = session.read.table("update_employee")
    assert (
        df2.where(df2["employee_id"] == 1).select("age").collect()[0]["age"]
        == df_employee.where(df_employee["employee_id"] == 1).select("age").collect()[0]["age"] + 1
    )


def test_delete_table(cleanup_employee_df: BaseDataFrame, caplog):
    session = cleanup_employee_df.session
    if isinstance(
        session,
        (
            StandaloneSession,
            PySparkSession,
            SparkSession,
        ),
    ):
        pytest.skip("Engine doesn't support delete")
    df_employee = cleanup_employee_df
    df_employee.write.saveAsTable("delete_employee")
    df = session.read.table("delete_employee")
    assert isinstance(df, _BaseTable)
    delete_expr = df.delete(where=df["age"] > 28)
    result = delete_expr.execute()
    # Postgres, RedshiftSession and BigQuery don't support returning the number of affected rows
    if not isinstance(session, (PostgresSession, BigQuerySession, RedshiftSession)):
        assert result[0][0] == 4

    df2 = session.read.table("delete_employee")
    assert df2.collect() == [
        Row(employee_id=4, fname="Claire", lname="Littleton", age=27, store_id=2)
    ]


def test_merge_table_simple(cleanup_employee_df: BaseDataFrame, caplog):
    session = cleanup_employee_df.session
    if isinstance(
        session,
        (
            StandaloneSession,
            PySparkSession,
            RedshiftSession,
            SparkSession,
            DuckDBSession,
        ),
    ):
        pytest.skip("Engine doesn't support merge")
    df_employee = cleanup_employee_df
    df_employee.write.saveAsTable("merge_employee")
    df = session.read.table("merge_employee")
    assert isinstance(df, _BaseTable)

    df2 = session.createDataFrame(
        [
            (1, "Jack", "Shephard", 38, 2),
            (6, "Mary", "Sue", 21, 45),
        ],
        ["employee_id", "fname", "lname", "age", "store_id"],
    )

    merge_expr = df.merge(
        df2,
        condition=df.employee_id == df2.employee_id,
        clauses=[
            WhenMatched(condition=df.fname == df2.fname).update(
                set_={
                    "lname": df2.lname,
                    "age": df2.age,
                    "store_id": df2.store_id,
                }
            ),
            WhenNotMatched().insert(
                values={
                    "employee_id": df2.employee_id,
                    "fname": df2.fname,
                    "lname": df2.lname,
                    "age": df2.age,
                    "store_id": df2.store_id,
                }
            ),
        ],
    )
    result = merge_expr.execute()
    # Postgres and BigQuery don't support returning the number of affected rows
    if not isinstance(session, (PostgresSession, BigQuerySession)):
        if isinstance(session, SnowflakeSession):
            assert (result[0][0] + result[0][1]) == 2
        else:
            assert result[0][0] == 2

    df_merged = session.read.table("merge_employee")
    assert sorted(df_merged.collect()) == [
        Row(employee_id=1, fname="Jack", lname="Shephard", age=38, store_id=2),
        Row(employee_id=2, fname="John", lname="Locke", age=65, store_id=1),
        Row(employee_id=3, fname="Kate", lname="Austen", age=37, store_id=2),
        Row(employee_id=4, fname="Claire", lname="Littleton", age=27, store_id=2),
        Row(employee_id=5, fname="Hugo", lname="Reyes", age=29, store_id=100),
        Row(employee_id=6, fname="Mary", lname="Sue", age=21, store_id=45),
    ]


def test_merge_table(cleanup_employee_df: BaseDataFrame, merge_data, get_func, caplog):
    session = cleanup_employee_df.session
    col = get_func("col", session)
    expr = get_func("expr", session)
    lit = get_func("lit", session)
    if isinstance(
        session,
        (
            StandaloneSession,
            PySparkSession,
            RedshiftSession,
            SparkSession,
            DuckDBSession,
        ),
    ):
        pytest.skip("Engine doesn't support merge")

    if isinstance(session, DatabricksSession):
        uuid_func = "uuid"
    elif isinstance(session, PostgresSession):
        uuid_func = "gen_random_uuid"
    elif isinstance(session, SnowflakeSession):
        uuid_func = "uuid_string"
    elif isinstance(session, BigQuerySession):
        uuid_func = "generate_uuid"
    else:
        pytest.skip("Cannot generate uuids in this engine")

    data, schema = merge_data
    df_employee = session.createDataFrame(data, schema)
    df_employee.write.saveAsTable("merge_employee")
    df = session.read.table("merge_employee")
    assert isinstance(df, _BaseTable)

    start_date = "2024-01-01"
    end_date = "2900-01-01"
    df2 = session.createDataFrame(
        [
            (1, "Jack", "Shephard", 38, 2),
            (6, "Mary", "Sue", 21, 45),
        ],
        ["employee_id", "fname", "lname", "age", "store_id"],
    )

    source = df2.select(
        expr(f"{uuid_func}()").alias("s_id"),
        col("employee_id").alias("employee_id"),
        col("fname").alias("fname"),
        col("lname").alias("lname"),
        col("age").alias("age"),
        col("store_id").alias("store_id"),
        lit(end_date).alias("end_date"),
    )

    updates = (
        source.alias("src")
        .join(
            df.alias("tg"),
            on=(col("tg.employee_id") == col("src.employee_id")),
        )
        .where(
            f"tg.end_date = '{end_date}' "
            f"AND ("
            f"src.fname <> tg.fname OR src.lname <> tg.lname OR src.age <> tg.age OR src.store_id <> tg.store_id"
            f")"
        )
    )

    staging = updates.select(
        lit(0).alias("__key__"),
        col("src.*"),
        lit(start_date).cast("date").alias("start_date"),
    ).unionByName(
        source.alias("src").select(
            lit(1).alias("__key__"),
            col("src.*"),
            lit(start_date).cast("date").alias("start_date"),
        )
    )

    merge_expr = df.alias("tg").merge(
        staging.alias("st"),
        condition=(col("st.employee_id") == col("tg.employee_id")) & (col("st.__key__") == lit(1)),
        clauses=[
            WhenMatched(
                condition=(col("tg.end_date") == lit(end_date))
                & (
                    (col("st.fname") != col("tg.fname"))
                    | (col("st.lname") != col("tg.lname"))
                    | (col("st.age") != col("tg.age"))
                    | (col("st.store_id") != col("tg.store_id"))
                )
            ).update(
                set_={
                    "end_date": col("st.start_date"),
                }
            ),
            WhenNotMatched().insert(
                values={
                    "s_id": col("st.s_id"),
                    "employee_id": col("st.employee_id"),
                    "fname": col("st.fname"),
                    "lname": col("st.lname"),
                    "age": col("st.age"),
                    "store_id": col("st.store_id"),
                    "start_date": lit(start_date),
                    "end_date": lit(end_date),
                }
            ),
        ],
    )

    result = merge_expr.execute()
    if not isinstance(session, (PostgresSession, BigQuerySession)):
        if isinstance(session, SnowflakeSession):
            assert (result[0][0] + result[0][1]) == 3
        else:
            assert result[0][0] == 3

    df_merged = session.read.table("merge_employee").select(
        "employee_id", "fname", "lname", "age", "store_id", "start_date", "end_date"
    )

    assert df_merged.count() == 7
    assert sorted(df_merged.collect()) == [
        Row(
            employee_id=1,
            fname="Jack",
            lname="Shephard",
            age=37,
            store_id=1,
            start_date=datetime.date(1999, 1, 1),
            end_date=datetime.date(2024, 1, 1),
        ),
        Row(
            employee_id=1,
            fname="Jack",
            lname="Shephard",
            age=38,
            store_id=2,
            start_date=datetime.date(2024, 1, 1),
            end_date=datetime.date(2900, 1, 1),
        ),
        Row(
            employee_id=2,
            fname="John",
            lname="Locke",
            age=65,
            store_id=1,
            start_date=datetime.date(1999, 1, 1),
            end_date=datetime.date(2900, 1, 1),
        ),
        Row(
            employee_id=3,
            fname="Kate",
            lname="Austen",
            age=37,
            store_id=2,
            start_date=datetime.date(1999, 1, 1),
            end_date=datetime.date(2900, 1, 1),
        ),
        Row(
            employee_id=4,
            fname="Claire",
            lname="Littleton",
            age=27,
            store_id=2,
            start_date=datetime.date(1999, 1, 1),
            end_date=datetime.date(2900, 1, 1),
        ),
        Row(
            employee_id=5,
            fname="Hugo",
            lname="Reyes",
            age=29,
            store_id=100,
            start_date=datetime.date(1999, 1, 1),
            end_date=datetime.date(2900, 1, 1),
        ),
        Row(
            employee_id=6,
            fname="Mary",
            lname="Sue",
            age=21,
            store_id=45,
            start_date=datetime.date(2024, 1, 1),
            end_date=datetime.date(2900, 1, 1),
        ),
    ]



from sqlframe.base.mixins.table_mixins import (
    DeleteSupportMixin,
    MergeSupportMixin,
    UpdateSupportMixin,
)
from sqlframe.base.table import (
    WhenMatched,
    WhenNotMatched,
    WhenNotMatchedBySource,
    _BaseTable,
)
from sqlframe.databricks.dataframe import DatabricksDataFrame


class DatabricksTable(
    DatabricksDataFrame,
    UpdateSupportMixin["DatabricksDataFrame"],
    DeleteSupportMixin["DatabricksDataFrame"],
    MergeSupportMixin["DatabricksDataFrame"],
    _BaseTable["DatabricksDataFrame"],
):
    _df = DatabricksDataFrame
    _merge_supported_clauses = [WhenMatched, WhenNotMatched, WhenNotMatchedBySource]
    _merge_support_star = True



import sys
import typing as t
from enum import IntEnum
from uuid import uuid4

from sqlglot import exp
from sqlglot.expressions import _to_s
from sqlglot.helper import object_to_dict

from sqlframe.base.dataframe import DF, SESSION, BaseDataFrame

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if t.TYPE_CHECKING:
    from sqlframe.base._typing import ColumnOrLiteral
    from sqlframe.base.column import Column
    from sqlframe.base.types import Row


class Clause(IntEnum):
    UPDATE = 1
    UPDATE_ALL = 2
    DELETE = 3
    INSERT = 4
    INSERT_ALL = 5


class MergeClause:
    def __init__(
        self,
        clause_type: Clause,
        condition: t.Optional[t.Union[str, t.List[str], "Column", t.List["Column"], bool]] = None,
        assignments: t.Optional[
            t.Dict[t.Union["Column", str], t.Union["Column", "ColumnOrLiteral", exp.Expression]]
        ] = None,
        matched: bool = True,
        by_source: bool = False,
    ):
        self.clause_type = clause_type
        self.condition = condition
        self.assignments = assignments
        self.matched = matched
        self.by_source = by_source


class WhenMatched:
    def __init__(
        self,
        condition: t.Optional[t.Union[str, t.List[str], "Column", t.List["Column"], bool]] = None,
    ):
        self._condition = condition
        self._clause: t.Union[MergeClause, None] = None

    def update(
        self,
        set_: t.Dict[t.Union["Column", str], t.Union["Column", "ColumnOrLiteral", exp.Expression]],
    ) -> Self:
        if self._clause:
            clause = self._clause.clause_type.name.lower()
            raise ValueError(f"WhenMatched already has an '{clause}' clause")
        self._clause = MergeClause(
            Clause.UPDATE,
            self._condition,
            {k: v for k, v in set_.items()},
            matched=True,
            by_source=False,
        )
        return self

    def update_all(self) -> Self:
        if self._clause:
            clause = self._clause.clause_type.name.lower()
            raise ValueError(f"WhenMatched already has an '{clause}' clause")
        self._clause = MergeClause(
            Clause.UPDATE_ALL,
            self._condition,
            {},
            matched=True,
            by_source=False,
        )
        return self

    def delete(self) -> Self:
        if self._clause:
            clause = self._clause.clause_type.name.lower()
            raise ValueError(f"WhenMatched already has an '{clause}' clause")
        self._clause = MergeClause(Clause.DELETE, self._condition, matched=True, by_source=False)
        return self

    @property
    def clause(self):
        return self._clause


class WhenNotMatched:
    def __init__(
        self,
        condition: t.Optional[t.Union[str, t.List[str], "Column", t.List["Column"], bool]] = None,
    ):
        self._condition = condition
        self._clause: t.Union[MergeClause, None] = None

    def insert(
        self,
        values: t.Dict[
            t.Union["Column", str], t.Union["Column", "ColumnOrLiteral", exp.Expression]
        ],
    ) -> Self:
        if self._clause:
            clause = self._clause.clause_type.name.lower()
            raise ValueError(f"WhenNotMatched already has an '{clause}' clause")
        self._clause = MergeClause(
            Clause.INSERT,
            self._condition,
            {k: v for k, v in values.items()},
            matched=False,
            by_source=False,
        )
        return self

    def insert_all(self) -> Self:
        if self._clause:
            clause = self._clause.clause_type.name.lower()
            raise ValueError(f"WhenNotMatched already has an '{clause}' clause")
        self._clause = MergeClause(
            Clause.INSERT_ALL,
            self._condition,
            {},
            matched=False,
            by_source=False,
        )
        return self

    @property
    def clause(self):
        return self._clause


class WhenNotMatchedBySource(object):
    def __init__(
        self,
        condition: t.Optional[t.Union[str, t.List[str], "Column", t.List["Column"], bool]] = None,
    ):
        self._condition = condition
        self._clause: t.Union[MergeClause, None] = None

    def update(
        self,
        set_: t.Dict[t.Union["Column", str], t.Union["Column", "ColumnOrLiteral", exp.Expression]],
    ) -> Self:
        if self._clause:
            clause = self._clause.clause_type.name.lower()
            raise ValueError(f"WhenNotMatchedBySource already has an '{clause}' clause")
        self._clause = MergeClause(
            Clause.UPDATE,
            self._condition,
            {k: v for k, v in set_.items()},
            matched=False,
            by_source=True,
        )
        return self

    def delete(self) -> Self:
        if self._clause:
            clause = self._clause.clause_type.name.lower()
            raise ValueError(f"WhenNotMatchedBySource already has an '{clause}' clause")
        self._clause = MergeClause(Clause.DELETE, self._condition, matched=False, by_source=True)
        return self

    @property
    def clause(self):
        return self._clause


class LazyExpression:
    def __init__(
        self,
        expression: exp.Expression,
        session: SESSION,
    ):
        self._expression = expression
        self._session = session

    def execute(self) -> t.List["Row"]:
        return self._session._collect(self._expression)

    @property
    def expression(self) -> exp.Expression:
        return self._expression

    def __str__(self) -> str:
        return self._expression.sql()

    def __repr__(self) -> str:
        return _to_s(self._expression)


class _BaseTable(BaseDataFrame, t.Generic[DF]):
    _df: t.Type[DF]

    def copy(self, **kwargs):
        kwargs["join_on_uuid"] = str(uuid4())
        return self._df(**object_to_dict(self, **kwargs))

    def __copy__(self):
        return self.copy()

    def table_copy(self):
        return self.__class__(**object_to_dict(self))

    def alias(self, name: str, **kwargs) -> Self:
        df = BaseDataFrame.alias(self, name, **kwargs)
        new_df = self.__class__(**object_to_dict(df))
        return new_df

    def update(
        self,
        set_: t.Dict[t.Union["Column", str], t.Union["Column", "ColumnOrLiteral", exp.Expression]],
        where: t.Optional[t.Union["Column", str, bool]] = None,
    ) -> LazyExpression:
        raise NotImplementedError()

    def merge(
        self,
        source: DF,
        condition: t.Union[str, t.List[str], "Column", t.List["Column"], bool],
        clauses: t.Iterable[t.Union[WhenMatched, WhenNotMatched, WhenNotMatchedBySource]],
    ) -> LazyExpression:
        raise NotImplementedError()

    def delete(
        self,
        where: t.Optional[t.Union["Column", str, bool]] = None,
    ) -> LazyExpression:
        raise NotImplementedError()



from sqlframe.base.mixins.table_mixins import (
    DeleteSupportMixin,
    MergeSupportMixin,
    UpdateSupportMixin,
)
from sqlframe.base.table import (
    WhenMatched,
    WhenNotMatched,
    _BaseTable,
)
from sqlframe.snowflake.dataframe import SnowflakeDataFrame


class SnowflakeTable(
    SnowflakeDataFrame,
    UpdateSupportMixin["SnowflakeDataFrame"],
    DeleteSupportMixin["SnowflakeDataFrame"],
    MergeSupportMixin["SnowflakeDataFrame"],
    _BaseTable["SnowflakeDataFrame"],
):
    _df = SnowflakeDataFrame
    _merge_supported_clauses = [WhenMatched, WhenNotMatched]
    _merge_support_star = False
